@@ -3,17 +3,17 @@ import inspect
 import logging
 import pkgutil
 import warnings
-from typing import Text, List, Dict, Any, Type, Union, Callable, Optional
+from typing import Text, List, Dict, Any, Type, Union, Callable, Optional, Set
+from collections import namedtuple
 import typing
 import types
+import sys
+import os
 
-from rasa_sdk.interfaces import Tracker, ActionNotFoundException
+from rasa_sdk.interfaces import Tracker, ActionNotFoundException, Action
 
 from rasa_sdk import utils
 
-if typing.TYPE_CHECKING:
-    from rasa_sdk.interfaces import Action
-
 logger = logging.getLogger(__name__)
 
 
@@ -142,19 +142,29 @@ class CollectingDispatcher:
         self.utter_message(image=image, **kwargs)
 
 
+TimestampModule = namedtuple("TimestampModule", ["timestamp", "module"])
+
+
 class ActionExecutor:
-    def __init__(self):
+    def __init__(self) -> None:
         self.actions = {}
+        self._modules: Dict[Text, TimestampModule] = {}
+        self._loaded: Set[Type] = set()
 
     def register_action(self, action: Union[Type["Action"], "Action"]) -> None:
-        from rasa_sdk.interfaces import Action
-
         if inspect.isclass(action):
             if action.__module__.startswith("rasa."):
                 logger.warning(f"Skipping built in Action {action}.")
                 return
             else:
+                # Check if this class has already been loaded in the past.
+                if action in self._loaded:
+                    return
+
+                # Mark the class as "loaded"
+                self._loaded.add(action)
                 action = action()
+
         if isinstance(action, Action):
             self.register_function(action.name(), action.run)
         else:
@@ -165,7 +175,6 @@ class ActionExecutor:
             )
 
     def register_function(self, name: Text, f: Callable) -> None:
-        logger.info(f"Registered function for '{name}'.")
         valid_keys = utils.arguments_of(f)
         if len(valid_keys) < 3:
             raise Exception(
@@ -175,52 +184,150 @@ class ActionExecutor:
                 f"tracker, domain. Your function accepts only {len(valid_keys)} "
                 "parameters."
             )
+
+        if name in self.actions:
+            logger.info(f"Re-registered function for '{name}'.")
+        else:
+            logger.info(f"Registered function for '{name}'.")
+
         self.actions[name] = f
 
     def _import_submodules(
         self, package: Union[Text, types.ModuleType], recursive: bool = True
     ) -> None:
-        """ Import all submodules of a module, recursively, including
-        subpackages
+        """Import a module, or a package and its submodules recursively.
 
-        :param package: package (name or actual module)
-        :type package: str | module
-        :rtype: dict[str, types.ModuleType]
+        Args:
+            package: Package or module name, or an already loaded Python module.
+            recursive: If `True`, and `package` is a package, import all of its
+                sub-packages as well.
         """
         if isinstance(package, str):
-            package = importlib.import_module(package)
+            package = self._import_module(package)
+
         if not getattr(package, "__path__", None):
             return
 
-        results = {}
         for loader, name, is_pkg in pkgutil.walk_packages(package.__path__):
             full_name = package.__name__ + "." + name
-            results[full_name] = importlib.import_module(full_name)
+            self._import_module(full_name)
+
             if recursive and is_pkg:
                 self._import_submodules(full_name)
 
-    def register_package(self, package: Union[Text, types.ModuleType]) -> None:
-        from rasa_sdk.interfaces import Action
+    def _import_module(self, name: Text) -> types.ModuleType:
+        """Import a Python module. If possible, register the file where it came from.
 
+        Args:
+            name: The module's name using Python's module path syntax.
+
+        Returns:
+            The loaded module.
+        """
+        module = importlib.import_module(name)
+
+        module_file = getattr(module, "__file__", None)
+        if module_file:
+            # If the module we're importing is a namespace package (a package
+            # without __init__.py), then there's nothing to watch for the
+            # package itself.
+            timestamp = os.path.getmtime(module_file)
+            self._modules[module_file] = TimestampModule(timestamp, module)
+
+        return module
+
+    def register_package(self, package: Union[Text, types.ModuleType]) -> None:
+        """Register all the `Action` subclasses contained in a Python module or
+        package.
+
+        If an `ImportError` is raised when loading the module or package, the
+        action server is stopped with exit code 1.
+
+        Args:
+            package: Module or package name, or an already loaded Python module.
+        """
         try:
             self._import_submodules(package)
         except ImportError:
             logger.exception(f"Failed to register package '{package}'.")
+            sys.exit(1)
+
+        self._register_all_actions()
+
+    def _register_all_actions(self) -> None:
+        """Scan for all user subclasses of `Action`, and register them.
+        """
+        import inspect
 
         actions = utils.all_subclasses(Action)
 
         for action in actions:
-            meta = action.__dict__.get("Meta", False)
-            abstract = getattr(meta, "abstract", False)
             if (
                 not action.__module__.startswith("rasa_core.")
                 and not action.__module__.startswith("rasa.")
                 and not action.__module__.startswith("rasa_sdk.")
                 and not action.__module__.startswith("rasa_core_sdk.")
-                and not abstract
+                and not inspect.isabstract(action)
             ):
                 self.register_action(action)
 
+    def _find_modules_to_reload(self) -> Dict[Text, TimestampModule]:
+        """Finds all Python modules that should be reloaded by checking their
+        files' timestamps.
+
+        Returns:
+            Dictionary containing file paths, new timestamps and Python modules
+            that should be reloaded.
+        """
+
+        to_reload = {}
+
+        for path, (timestamp, module) in self._modules.items():
+            try:
+                new_timestamp = os.path.getmtime(path)
+            except OSError:
+                # Ignore missing files
+                continue
+
+            if new_timestamp > timestamp:
+                to_reload[path] = TimestampModule(new_timestamp, module)
+
+        return to_reload
+
+    def reload(self) -> None:
+        """Reload all Python modules that have been loaded in the past.
+
+        To check if a module should be reloaded, the file's last timestamp is
+        checked against the current one. If the current last-modified timestamp
+        is larger than the previous one, the module is re-imported using
+        `importlib.reload`.
+
+        If one or more modules are reloaded during this process, the entire
+        `Action` class hierarchy is scanned again to see what new classes can
+        be registered.
+        """
+        to_reload = self._find_modules_to_reload()
+        any_module_reloaded = False
+
+        for path, (timestamp, module) in to_reload.items():
+            try:
+                new_module = importlib.reload(module)
+                self._modules[path] = TimestampModule(timestamp, new_module)
+                logger.info(
+                    f"Reloaded module/package: '{module.__name__}' "
+                    f"(file: '{os.path.relpath(path)}')"
+                )
+                any_module_reloaded = True
+            except (SyntaxError, ImportError):
+                logger.exception(
+                    f"Error while reloading module/package: '{module.__name__}' "
+                    f"(file: '{os.path.relpath(path)}'):"
+                )
+                logger.info("Please fix the error(s) in the Python file and try again.")
+
+        if any_module_reloaded:
+            self._register_all_actions()
+
     @staticmethod
     def _create_api_response(
         events: List[Dict[Text, Any]], messages: List[Dict[Text, Any]]
